# Copyright (C) 2026 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Optional
from pathlib import Path

from .HeavyIrObject import HeavyIrObject
from .HeavyGraph import HeavyGraph

from hvcc.generators.ir2c.SignalNamHeader import ModelNet, load_nam_file


class HIrNam(HeavyIrObject):
    """ __nam~f
    """

    def __init__(
        self,
        obj_type: str,
        args: Optional[Dict] = None,
        graph: Optional[HeavyGraph] = None,
        annotations: Optional[Dict] = None
    ) -> None:
        assert obj_type == "__nam~f"

        # load the nam file to retrieve the model type
        assert args is not None
        model_net, _, _ = load_nam_file(Path(args["nam"]))

        # overload the object type based on the model
        if model_net == ModelNet.Nano:
            obj_type = "__nam_nano~f"
        elif model_net == ModelNet.Feather:
            obj_type = "__nam_feather~f"
        elif model_net == ModelNet.Lite:
            obj_type = "__nam_lite~f"
        elif model_net == ModelNet.Standard:
            obj_type = "__nam_standard~f"

        super().__init__(obj_type, args=args, graph=graph, annotations=annotations)



# Copyright (C) 2022-2025 Daniel Billotte, Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Optional

from .HeavyIrObject import HeavyIrObject
from .HeavyException import HeavyException
from .HeavyGraph import HeavyGraph


class HIrExpr(HeavyIrObject):
    """ This needs little logic as the expression is forwarded to ir2c
    """

    def __init__(
        self,
        obj_type: str,
        args: Optional[Dict] = None,
        graph: Optional[HeavyGraph] = None,
        annotations: Optional[Dict] = None
    ) -> None:
        super().__init__(obj_type, args=args, graph=graph,
                         num_inlets=args["num_inlets"] if args is not None else 1,
                         num_outlets=1,
                         annotations=annotations)

    def inlet_requires_signal(self, inlet_index: int = 0) -> bool:
        """ Overriding HeavyIrObject to deal with variable number of inlets
        """
        if inlet_index >= self.num_inlets:
            raise HeavyException(f"inlet_index: {inlet_index} is greater than number of inlets")
        return True



import json

from copy import deepcopy
from typing import Any, Dict, Optional

from hvcc.core.hv2ir.HeavyLangObject import HeavyLangObject
from hvcc.types.compiler import ExternParams


def filter_match(
    set: Dict,
    key: str,
    match: str,
    key_exclude: Optional[str] = None,
    match_exclude: Optional[str] = None
) -> filter:
    if (key_exclude is not None and match_exclude is not None):
        return filter(lambda x: x.get(key, '') == match and x.get(key_exclude, '') != match_exclude, set)
    else:
        return filter(lambda x: x.get(key, '') == match, set)


def verify_param_exists(name: str, original_name: str, components: Dict, input: bool = True) -> Any:
    for comp in components:

        # Dealing with the cvouts the way we have it set up is really annoying
        if comp['component'] == 'CVOuts':

            if name == comp['name']:
                if input:
                    raise TypeError(
                        f'Parameter "{original_name}" cannot be used as an {"input" if input else "output"}')
                return
        else:
            variants = [mapping['name'].format_map(
                {'name': comp['name']}) for mapping in comp['mapping']]
            if name in variants:
                if input and comp['direction'] == 'output' or not input and comp['direction'] == 'input':
                    raise TypeError(
                        f'Parameter "{original_name}" cannot be used as an {"input" if input else "output"}')
                return

    raise NameError(f'Unknown parameter "{original_name}"')


def verify_param_direction(name: str, components: Dict) -> bool:
    for comp in components:
        if comp['component'] == 'CVOuts':
            if name == comp['name']:
                return True
        else:
            variants = [mapping['name'].format_map(
                {'name': comp['name']}) for mapping in comp['mapping']]
            if name in variants:
                return True
    return False


def get_root_component(variant: str, original_name: str, components: Dict) -> str:
    for comp in components:
        if comp['component'] == 'CVOuts':
            if variant == comp['name']:
                return variant
        else:
            variants = [mapping['name'].format_map(
                {'name': comp['name']}) for mapping in comp['mapping']]
            if variant in variants:
                return comp['name']
    raise NameError(f'Unknown parameter "{original_name}"')


def get_component_mapping(
    component_variant: str,
    original_name: str,
    component: Dict,
    components: Dict
) -> Dict:
    for variant in component['mapping']:
        if component['component'] == 'CVOuts':
            stripped = variant['name'].format_map({'name': ''})
            if stripped in component['name']:
                return variant
        elif variant['name'].format_map({'name': component['name']}) == component_variant:
            return variant
    raise NameError(f'Unknown parameter "{original_name}"')


def verify_param_used(
    component: Dict,
    params_in: Dict,
    params_out: Dict,
    params_in_original_name: Dict,
    params_out_original_name: Dict,
    components: Dict
) -> bool:
    # Exclude parents, since they don't have 1-1 i/o mapping
    if component.get('is_parent', False):
        return True
    for param in {**params_in, **params_out}:
        root = get_root_component(
            param, ({**params_in_original_name, **params_out_original_name})[param], components)
        if root == component['name']:
            return True
    return False


def de_alias(
    name: str,
    aliases: Dict,
    components: Dict
) -> str:
    low = name.lower()
    # simple case
    if low in aliases:
        return aliases[low]
    # aliased variant
    potential_aliases = list(filter(lambda x: x in low, aliases))
    for alias in potential_aliases:
        try:
            target_component = list(filter_match(
                components, 'name', aliases[alias]))[0]
            # The CVOuts setup really bothers me
            if target_component['component'] != 'CVOuts':
                for mapping in target_component['mapping']:
                    if mapping['name'].format_map({'name': alias}) == low:
                        return mapping['name'].format_map({'name': aliases[alias]})
        except IndexError:
            # No matching alias from filter
            pass
    # otherwise, it's a direct parameter or unkown one
    return low


def parse_parameters(
    parameters: ExternParams,
    components: Dict,
    aliases: Dict,
    object_name: str
) -> Dict:
    """
    Parses the `parameters` passed from hvcc and generates getters and setters
    according to the info in `components`. The `aliases` help disambiguate parameters
    and the `object_name` sets the identifier for the generated Daisy hardware class,
    in this case `hardware`.
    """

    # Verify that the params are valid and remove unused components
    replacements: Dict = {}
    params_in: Dict = {}
    params_in_original_names: Dict = {}
    for key, recv in parameters.inParam:
        de_aliased = de_alias(key, aliases, components)
        params_in[de_aliased] = recv
        params_in_original_names[de_aliased] = key

    params_out = {}
    params_out_original_names = {}
    for key, msg in parameters.outParam:
        de_aliased = de_alias(key, aliases, components)
        params_out[de_aliased] = msg
        params_out_original_names[de_aliased] = key

    [verify_param_exists(key, params_in_original_names[key],
                         components, input=True) for key in params_in]
    [verify_param_exists(key, params_out_original_names[key],
                         components, input=False) for key in params_out]

    for i in range(len(components) - 1, -1, -1):
        if not verify_param_used(
                components[i], params_in, params_out,
                params_in_original_names, params_out_original_names,
                components):
            components.pop(i)

    out_idx = 0

    replacements['parameters'] = []
    replacements['output_parameters'] = []
    replacements['loop_write_in'] = []
    replacements['callback_write_out'] = []
    replacements['loop_write_out'] = []
    replacements['hook_write_out'] = []
    replacements['callback_write_in'] = []

    for param_name, param in params_in.items():
        root = get_root_component(
            param_name, params_in_original_names[param_name], components)
        component = list(filter_match(components, 'name', root))[0]
        param_struct = {
            "hash_enum": params_in_original_names[param_name], 'name': root, 'type': component['component'].upper()}
        replacements['parameters'].append(param_struct)
        mapping = get_component_mapping(
            param_name, params_in_original_names[param_name], component, components)

        write_location = 'callback_write_in' if mapping.get(
            'where', 'callback') == 'callback' else 'loop_write_in'
        component_info = deepcopy(component)
        component_info['name'] = root
        component_info['class_name'] = object_name
        # A bit of a hack to get cv_1, etc to be written as CV_1
        component_info['name_upper'] = root.upper()
        component_info['value'] = f'output_data[{out_idx}]'
        component_info['default_prefix'] = component.get(
            "default_prefix", '') if component.get('default', False) else ''
        process = mapping["get"].format_map(component_info)

        replacements[write_location].append(
            {"process": process, "bool": mapping.get('bool', False),
                "hash_enum": params_in_original_names[param_name]})

    for param_name, param in params_out.items():
        root = get_root_component(
            param_name, params_out_original_names[param_name], components)
        component = list(filter_match(components, 'name', root))[0]

        mapping = get_component_mapping(
            param_name, params_out_original_names[param_name], component, components)

        default_prefix = component.get(
            "default_prefix", '') if component.get('default', False) else ''

        write_locations = {'callback': 'callback_write_out',
                           'loop': 'loop_write_out',
                           'hook': 'hook_write_out'}

        write_location = write_locations.get(mapping.get('where', 'callback'), 'callback_write_out')

        param_struct = {
            'hash_enum': params_out_original_names[param_name],
            'index': out_idx, 'name': param_name,
            'hook': write_location == 'hook_write_out'}
        replacements['output_parameters'].append(param_struct)

        component_info = deepcopy(component)
        component_info['hash_enum'] = params_out_original_names[param_name]
        component_info['name'] = root
        component_info['class_name'] = object_name
        component_info['value'] = f'output_data[{out_idx}]' if \
            write_location != 'hook_write_out' else 'sig'
        component_info['default_prefix'] = default_prefix
        write = mapping["set"].format_map(component_info)

        replacements[write_location].append(
            {"name": param_name,
             "process": write,
             "bool": mapping.get('bool', False),
             "value": component_info['value']})

        out_idx += 1

    replacements['output_comps'] = len(replacements['output_parameters'])

    return replacements


def display_parameters(description_file: str) -> dict[str, str]:
    """
    Optional list of externed parameters that will be filtered out of the hardware configuration.
    These will be passed on to the sendHook callback and can be used by the Display function.
    """

    with open(description_file, 'rb') as file:
        daisy_description = json.load(file)

        try:
            params = {
                param: "0x{0:X}".format(HeavyLangObject.get_hash(param))
                for param in daisy_description['display']['params']
            }
        except KeyError:
            return {}

        return params


def display_processor(description_file: Optional[str] = None) -> str:
    """
    Try to load display processing code from file.
    """

    if description_file is None:
        raise ValueError('description_file not provided')

    with open(description_file, 'rb') as file:
        daisy_description = json.load(file)

        process_file = daisy_description['display']['process_file']
        with open(process_file, 'rb') as f:
            return f.read().decode('utf-8')



import json

from pathlib import Path

import jinja2

from hvcc.types.GUI import Canvas, Comment, GraphRoot, Graph, GUIObjects


def open_gui_json(
    patch_name: str,
    c_src_dir: Path,
) -> GraphRoot:

    # load GUI from json file
    gui_json_path = Path(c_src_dir, "../ir/", f"{patch_name}.heavy.gui.json")
    with open(gui_json_path, "r") as f:
        gui_json = GraphRoot(**json.load(f))

    return gui_json


def nanovg_render(
    patch_name: str,
    c_src_dir: Path,
    env: jinja2.Environment,
    recv_list: list,
    send_list: list
) -> tuple[
    GraphRoot,
    dict[str, list[str]],
    list[str]
]:
    """ Generate nanovg components from the GUI json
    """
    gui_json = open_gui_json(patch_name, c_src_dir)

    # widget overview
    widgets: dict[str, list[str]] = {
        "graph": [],
        "canvas": [],
        "comment": [],
        "bang": [],
        "toggle": [],
        "vradio": [],
        "hradio": [],
        "vslider": [],
        "hslider": [],
        "knob": [],
        "number": [],
        "float": []
    }

    # render gui objects
    gui_objects_render = []

    def generate_gui_objects(graphs: list[Graph], objects: list[GUIObjects], parent: str):
        for w in objects:
            widgets[w.type].append(w.id if isinstance(w, (Canvas, Comment)) else w.parameter)

        gui_objects_render.append(env.get_template("gui_objects.cpp").render(
            parent=parent,
            gui_objects=objects,
            receivers=recv_list,
            senders=send_list
        ))

        for graph in graphs:
            widgets["graph"].append(graph.id)

            gui_objects_render.append(
                f"""
    // subpatch
    {graph.id} = new PDSubpatch({parent});
    {graph.id}->setSize({graph.gop_size.x} * scaleFactor, {graph.gop_size.y} * scaleFactor);
    {graph.id}->setAbsolutePos({graph.position.x} * scaleFactor, {graph.position.y} * scaleFactor);
    {parent}->addManagedChild({graph.id});
                """
                    )
            generate_gui_objects(graph.graphs, graph.objects, graph.id)

    generate_gui_objects(gui_json.graphs, gui_json.objects, "mainPatch")

    return gui_json, widgets, gui_objects_render



import jinja2
import json

from importlib import resources
from typing import Optional
from pathlib import Path


def map_load(pair: list, json_defs_file: Path):
    """
    Helper for loading and processing the definitions, component list, etc
    """

    # load the default components
    comp_string = Path(Path(__file__).parent, json_defs_file)
    component_defs = json.loads(comp_string.read_bytes())

    pair[1]['name'] = pair[0]

    # the default if it exists
    component = component_defs.get(pair[1]['component'], None)
    if (component):
        # copy component defs into the def
        # TODO this should be recursive for object structures..
        for k in component:
            if k not in pair[1]:
                pair[1][k] = component[k]
    else:
        raise Exception(f'unknown component "{pair[1]["component"]}"')

    return pair[1]


def filter_match(
    component_set: list,
    key: str,
    match: str,
    key_exclude: Optional[str] = None,
    match_exclude: bool = False
) -> filter:
    if (key_exclude is not None and match_exclude):
        return filter(lambda x: x.get(key, '') == match and x.get(key_exclude, '') != match_exclude, component_set)
    else:
        return filter(lambda x: x.get(key, '') == match, component_set)


def filter_matches(
    component_set: list,
    key: str,
    matches: list,
    key_exclude: Optional[str] = None,
    match_exclude: bool = False
) -> filter:
    if (key_exclude is not None and match_exclude):
        return filter(lambda x: x.get(key, '') in matches and x.get(key_exclude, '') != match_exclude, component_set)
    else:
        return filter(lambda x: x.get(key, '') in matches, component_set)


def filter_has(
    component_set: list,
    key: str,
    key_exclude: Optional[str] = None,
    match_exclude: bool = False
) -> filter:
    if (key_exclude is not None and match_exclude):
        return filter(lambda x: x.get(key, '') != '' and x.get(key_exclude, '') != match_exclude, component_set)
    else:
        return filter(lambda x: x.get(key, '') != '', component_set)


# filter out the components we need, then map them onto the init for that part
def filter_map_init(
    component_set: list,
    key: str,
    match: str,
    key_exclude: Optional[str] = None,
    match_exclude: bool = False
) -> str:
    filtered = filter_match(component_set, key, match, key_exclude=key_exclude, match_exclude=match_exclude)
    return "\n    ".join(map(lambda x: x['map_init'].format_map(x), filtered))


# this function is not used anywhere?
def filter_map_set(
    component_set: list,
    key: str,
    match: str,
    key_exclude: Optional[str] = None,
    match_exclude: bool = False
) -> str:
    filtered = filter_match(component_set, key, match, key_exclude=key_exclude, match_exclude=match_exclude)
    return "\n    ".join(
        map(lambda x: x['mapping'][0]['set'].format_map(x['mapping'][0]['name'].format_map(x)), filtered)
    )


def filter_map_ctrl(
    component_set: list,
    key: str,
    matches: list,
    init_key: str,
    key_exclude: Optional[str] = None,
    match_exclude: bool = False
) -> str:
    component_filt = filter_matches(component_set, key, matches, key_exclude=key_exclude, match_exclude=match_exclude)
    component_map = map(lambda x, i: {**x, 'i': i}, component_filt, range(1000))
    return "\n    ".join(map(lambda x: x[init_key].format_map(x), component_map))


# filter out the components with a certain field, then fill in the template
def filter_map_template(
    component_set: list,
    name: str,
    key_exclude: Optional[str] = None,
    match_exclude: bool = False
) -> str:
    filtered = filter_has(component_set, name, key_exclude=key_exclude, match_exclude=match_exclude)
    return "\n    ".join(map(lambda x: x[name].format_map(x), filtered))


def flatten_pin_dicts(comp: dict) -> dict:
    newcomp = {}
    for key, val in comp.items():
        if (isinstance(val, dict) and key == 'pin'):
            for subkey, subval in val.items():
                newcomp[f'{key}_{subkey}'] = subval
        else:
            newcomp[key] = val

    return newcomp


def flatten_index_dicts(comp: dict) -> dict:
    newcomp = {}
    for key, val in comp.items():
        if (isinstance(val, dict) and key == 'index'):
            for subkey, subval in val.items():
                newcomp[f'{key}_{subkey}'] = subval
        else:
            newcomp[key] = val

    return newcomp


def generate_header(board_description_dict: dict) -> 'tuple[str, dict]':
    """
    Generate a C++ Daisy board header from a dictionary board description.

    Returns a tuple containing the board
    header as a string and an information dictionary.

    The dictionary provides sufficient information to
    generate interface code, including component getters
    and setters, audio channel count, etc.
    """

    target = board_description_dict

    # flesh out target components:
    components = target.get('components', {})
    parents = target.get('parents', {})

    for key in parents:
        parents[key]['is_parent'] = True
    components.update(parents)

    seed_defs = Path("resources", 'component_defs.json')
    patchsm_defs = Path("resources", 'component_defs_patchsm.json')
    petalsm_defs = Path("resources", 'component_defs_petalsm.json')
    definitions = {'seed': seed_defs, 'patch_sm': patchsm_defs, 'petal_125b_sm': petalsm_defs}
    som = target.get('som', 'seed')

    try:
        json_defs_file = definitions[som]
    except KeyError:
        raise NameError(f'Unkown som "{som}"')

    # alphabetize by component name
    components = sorted(components.items(), key=lambda x: x[1]['component'])
    components = list(map(lambda p: map_load(p, json_defs_file), components))

    # flatten pin dicts into multiple entries
    # e.g. "pin": {"a": 12} => "pin_a": 12
    components = [flatten_pin_dicts(comp) for comp in components]
    components = [flatten_index_dicts(comp) for comp in components]

    target['components'] = components
    if 'name' not in target:
        target['name'] = 'custom'

    if 'aliases' not in target:
        target['aliases'] = {}

    if 'display' in target:
        # apply defaults if not present in config
        target['display']['driver'] = target['display'].get('driver', "daisy::SSD130x4WireSpi128x64Driver")
        target['display']['config'] = target['display'].get('config', [])
        target['display']['dim'] = target['display'].get('dim', [128, 64])
        target['displayprocess'] = target['display'].get('process', '')

    replacements = {}
    replacements['name'] = target['name']
    replacements['som'] = som
    replacements['external_codecs'] = target.get('external_codecs', [])

    som_classes = {
        'seed': 'daisy::DaisySeed',
        'patch_sm': 'daisy::patch_sm::DaisyPatchSM',
        'petal_125b_sm': 'daisy::Petal125BSM',
    }

    replacements['som_class'] = som_classes.get(som, som_classes['seed'])

    replacements['display_conditional'] = ('#include "dev/oled_ssd130x.h"' if ('display' in target) else "")
    replacements['target_name'] = target['name']
    replacements['init'] = filter_map_template(components, 'init', key_exclude='default', match_exclude=True)

    replacements['analogcount'] = len(list(filter_matches(
        components, 'component', ['AnalogControl', 'AnalogControlBipolar', 'CD4051'],
        key_exclude='default', match_exclude=True)))

    replacements['init_single'] = filter_map_ctrl(
        components, 'component', ['AnalogControl', 'AnalogControlBipolar', 'CD4051'],
        'init_single', key_exclude='default', match_exclude=True)
    replacements['ctrl_init'] = filter_map_ctrl(
        components, 'component', ['AnalogControl', 'AnalogControlBipolar'],
        'map_init', key_exclude='default', match_exclude=True)

    comp_string = Path(Path(__file__).parent, json_defs_file)
    definitions_dict = json.loads(comp_string.read_bytes())

    for name in definitions_dict:
        if name not in ('AnalogControl', 'AnalogControlBipolar', 'CD4051'):
            replacements[name] = filter_map_init(
                components, 'component', name, key_exclude='default', match_exclude=True)

    if 'display' in target:
        replacements['dispdec'] = f'daisy::OledDisplay<{target["display"]["driver"]}> display;'
        replacements['display'] = f"""
        daisy::OledDisplay<{target['display']['driver']}>::Config display_config;
        display_config.driver_config.transport_config.Defaults();
        {"".join(map(lambda x: x, target['display'].get('config', {})))}
        display.Init(display_config);
          display.Fill(0);
          display.Update();
        """
    else:
        replacements['display'] = ''

    if 'defines' in target:
        # deprecate old oopsy syntax at some point
        if target['defines'].get('HAS_MIDI') or target['defines'].get('OOPSY_TARGET_HAS_MIDI_INPUT'):
            target['has_midi'] = True
            replacements['midi'] = """daisy::MidiUartHandler midi;"""

    replacements['process'] = filter_map_template(
        components, 'process', key_exclude='default', match_exclude=True)
    replacements['loopprocess'] = filter_map_template(
        components, 'loopprocess', key_exclude='default', match_exclude=True)

    replacements['postprocess'] = filter_map_template(
        components, 'postprocess', key_exclude='default', match_exclude=True)
    replacements['displayprocess'] = filter_map_template(
        components, 'display', key_exclude='default', match_exclude=True)
    replacements['hidupdaterates'] = filter_map_template(
        components, 'updaterate', key_exclude='default', match_exclude=True)

    license_string = resources.files(__package__).joinpath('resources/LICENSE').read_text()
    replacements['license'] = '/*\n * ' + '\n * '.join([line for line in license_string.split('\n')]) + '\n */'

    component_declarations = list(filter(lambda x: not x.get('default', False), components))
    component_declarations = list(filter(lambda x: x.get('typename', '') != '', component_declarations))

    if len(component_declarations) > 0:
        replacements['comps'] = ";\n  ".join(
            map(lambda x: x['typename'].format_map(x) + ' ' + x['name'], component_declarations)
        ) + ';'

    non_class_declarations = list(filter(lambda x: 'non_class_decl' in x, component_declarations))

    if len(non_class_declarations) > 0:
        replacements['non_class_declarations'] = "\n".join(
            map(lambda x: x['non_class_decl'].format_map(x), non_class_declarations)
        )

    # env_opts = {"trim_blocks": True, "lstrip_blocks": True}

    # Ideally, this would be what we use, but we'll need to get the jinja PackageLoader class working
    # loader = jinja2.PackageLoader(__name__)
    # env = jinja2.Environment(loader=loader, **env_opts)
    # rendered_header = env.get_template('daisy.h').render(replacements)

    # This following works, but is really annoying
    header_str = Path(Path(__file__).parent, Path('templates', 'daisy.h')).read_text()
    header_env = jinja2.Environment(
        loader=jinja2.BaseLoader(),
        trim_blocks=True,
        lstrip_blocks=True
    ).from_string(header_str)

    rendered_header = header_env.render(replacements)
    rendered_header_lines = rendered_header.splitlines()
    rendered_header = "\n".join(line.rstrip() for line in rendered_header_lines) + "\n"

    # removing all unnecessary fields
    for comp in components:
        if 'map_init' in comp:
            del comp['map_init']
        if 'typename' in comp:
            del comp['typename']

    audio_info = target.get('audio', None)
    audio_channels = audio_info.get('channels', 2) if audio_info is not None else 2

    # This dictionary contains the necessary information to automatically (or manually)
    # write code to interface with the generated board
    board_info = {
        'name': target['name'],
        'components': components,
        'aliases': target['aliases'],
        'channels': audio_channels,
        'has_midi': target.get('has_midi', False),
        'displayprocess': target.get('displayprocess', '')
      }

    return rendered_header, board_info


def generate_header_from_file(description_file: Path) -> 'tuple[str, dict]':
    """
    Generate a C++ Daisy board header from a JSON description file.

    Returns a tuple containing the board
    header as a string and an information dictionary.

    The dictionary provides sufficient information to
    generate interface code, including component getters
    and setters, audio channel count, etc.
    """

    with open(description_file, 'rb') as file:
        daisy_description = json.load(file)

    return generate_header(daisy_description)


def generate_header_from_name(board_name: str) -> 'tuple[str, dict]':
    """
    Generate a C++ Daisy board header for an existing daisy board.

    Returns a tuple containing the board
    header as a string and an information dictionary.

    The dictionary provides sufficient information to
    generate interface code, including component getters
    and setters, audio channel count, etc.
    """

    try:
        description_file = Path('resources', f'{board_name}.json')
        daisy_description = Path(Path(__file__).parent, description_file)
        daisy_description_dict = json.loads(daisy_description.read_bytes())
    except FileNotFoundError:
        raise FileNotFoundError(f'Unknown Daisy board   "{board_name}"')

    return generate_header(daisy_description_dict)



# Heavy Compiler Collection
# Copyright (C) 2024 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

from enum import IntEnum
from typing import Dict, Literal, List, Optional, Tuple, Union
from typing_extensions import Self
from pydantic import BaseModel, HttpUrl, model_validator


DaisyBoards = Literal['pod', 'petal', 'patch', 'patch_init', 'field']
DaisyBootloader = Literal['BOOT_NONE', 'BOOT_SRAM', 'BOOT_QSPI']
DPFFormats = Literal['lv2', 'lv2_sep', 'vst2', 'vst3', 'clap', 'jack', 'dssi', 'ladspa', 'au']


class Daisy(BaseModel):
    board:          DaisyBoards = 'pod'
    board_file:     Optional[str] = None
    libdaisy_path:  Union[str, int] = 2
    usb_midi:       bool = False
    debug_printing: bool = False
    samplerate:     int = 48000
    blocksize:      Optional[int] = None
    linker_script:  str = ''
    bootloader:     Optional[DaisyBootloader] = None


class DPFUISize(BaseModel):
    width: int
    height: int


class DPFUIType(IntEnum):
    NONE = 0
    IMGUI = 1
    NANOVG = 2


class DPFPortGroups(BaseModel):
    input: Dict[str, Dict[str, Union[int, Tuple[int, bool]]]] = {}
    output: Dict[str, Dict[str, Union[int, Tuple[int, bool]]]] = {}


class DPF(BaseModel):
    dpf_path:           str = ""
    description:        Optional[str] = None
    makefile_dep:       List[str] = []
    enable_ui:          Optional[DPFUIType] = DPFUIType.NONE
    enable_modgui:      bool = False
    ui_size:            Optional[DPFUISize] = None
    midi_input:         bool = False
    midi_output:        bool = False
    port_groups:        Optional[DPFPortGroups] = None
    enumerators:        Optional[Dict[str, List[str]]] = None
    denormals:          bool = True
    version:            Optional[str] = None
    license:            Optional[str] = None
    maker:              Optional[str] = None
    brand_id:           Optional[str] = None
    brand_id_no_vst3:   Optional[bool] = False
    unique_id:          Optional[str] = None
    homepage:           Optional[HttpUrl] = None
    plugin_uri:         Optional[str] = None
    plugin_clap_id:     Optional[str] = None
    plugin_formats:     List[DPFFormats] = []
    lv2_info:           Optional[str] = None
    vst3_info:          Optional[str] = None
    clap_info:          List[str] = []

    @model_validator(mode='after')
    def au_required(self) -> Self:
        if 'au' in self.plugin_formats:
            if self.brand_id is None:
                raise ValueError("brand_id is required for au plugins")
            if self.unique_id is None:
                raise ValueError("unique_id is required for au plugins")
        return self


class Meta(BaseModel):
    name: Optional[str] = None
    nosimd: Optional[bool] = False
    daisy: Daisy = Daisy()
    dpf: DPF = DPF()
    external: Optional[Dict] = None



# Heavy Compiler Collection
# Copyright (C) 2025-2026 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

import argparse
import time

from typing import Optional
from pathlib import Path

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.interpreters.pd2gui.PdGUIParser import PdGUIParser
from hvcc.types.compiler import CompilerResp, CompilerNotif, CompilerMsg


class Colours:
    purple = "\033[95m"
    cyan = "\033[96m"
    dark_cyan = "\033[36m"
    blue = "\033[94m"
    green = "\033[92m"
    yellow = "\033[93m"
    red = "\033[91m"
    bold = "\033[1m"
    underline = "\033[4m"
    end = "\033[0m"


class pd2gui:

    @classmethod
    def compile(
        cls,
        pd_path: Path,
        ir_file: Path,
        search_paths: Optional[list] = None,
        verbose: bool = False
    ):
        tick = time.time()

        parser = PdGUIParser()
        if search_paths is not None:
            for p in search_paths:
                parser.add_absolute_search_directory(p)

        try:
            gui_graph, _ = parser.gui_from_file(pd_path)

            ir_dir = ir_file.parent
            if not ir_dir.exists():
                Path.mkdir(ir_dir)

            with open(ir_file, "w") as f:
                f.write(gui_graph.model_dump_json(indent=2) + "\n")

            return CompilerResp(
                stage="pd2gui",
                notifs=CompilerNotif(),
                in_dir=pd_path.parent,
                in_file=pd_path,
                out_dir=ir_dir,
                out_file=ir_file,
                compile_time=(time.time() - tick)
            )
        except Exception as e:
            return CompilerResp(
                stage="pd2gui",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                compile_time=(time.time() - tick)
            )


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Converts a Pd patch into the GUI format.")
    parser.add_argument(
        "pd_path",
        help="The Pd patch to convert to GUI.")
    parser.add_argument(
        "ir_dir",
        help="Directory to store generated GUI patches.")
    parser.add_argument(
        "-v",
        "--verbose",
        help="Show debugging information.",
        action="count")
    args = parser.parse_args()

    pd_path = Path(args.pd_path).expanduser().absolute()
    ir_dir = Path(args.ir_dir).expanduser().absolute()

    pd2gui.compile(
        pd_path=pd_path,
        ir_file=Path(ir_dir, f"{pd_path.stem}.heavy.gui.json"),
        search_paths=None,
        verbose=args.verbose)


if __name__ == "__main__":
    main()



# Copyright (C) 2022-2025 Daniel Billotte, Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from arpeggio import ParserPython, StrMatch     # type: ignore
from arpeggio import ZeroOrMore                 # type: ignore
from arpeggio import RegExMatch as regex        # type: ignore


class SuppressStrMatch(StrMatch):
    suppress = True


hide = SuppressStrMatch


def expr():     return none  # EOF                                                                  # noqa
def none():     return lor, ZeroOrMore(" ", lor)  # needed for logical `or`: `||`                   # noqa
def lor():      return land, ZeroOrMore("||", land)                                                 # noqa
def land():     return bor, ZeroOrMore("&&", bor)                                                   # noqa
def bor():      return xor, ZeroOrMore("|", xor)                                                    # noqa
def xor():      return band, ZeroOrMore("^", band)                                                  # noqa
def band():     return eq, ZeroOrMore("&", eq)                                                      # noqa
def eq():       return gtlte, ZeroOrMore(["==","!="], gtlte)                                        # noqa
def gtlte():    return gtlt, ZeroOrMore(["<=",">="], gtlt)                                          # noqa
def gtlt():     return shift, ZeroOrMore(["<",">"], shift)                                          # noqa
def shift():    return term, ZeroOrMore(["<<",">>"], term)                                          # noqa
def term():     return factor, ZeroOrMore(["+","-"], factor)    # RtoL                              # noqa
def factor():   return unary, ZeroOrMore(["*","/","%"], unary)                                      # noqa
def unary():    return [(["-","~","!"], unary), primary]        # RtoL                              # noqa
def primary():  return [number, var, func, group]                                                   # noqa
def number():   return [num_f, num_i]                                                               # noqa
def var():      return regex(r"\$[fisv]\d+")                                                        # noqa
def func():     return [                                                                            # noqa
                    (f_name, hide("("), expr, hide(")")),                                           # noqa
                    (f_name, hide("("), expr, hide(r"\,"), expr, hide(")")),                        # noqa
                    (f_name, hide("("), expr, hide(r"\,"), expr, hide(r"\,"), expr, hide(")"))      # noqa
                ]                                                                                   # noqa
def group():    return hide("("), expr, hide(")")                                                   # noqa
def num_f():    return regex(r"\d+\.\d+")  # prob need to cover exponential syntax, others?         # noqa
def num_i():    return regex(r"\d+")       # need to cover bin/octal/hex?                           # noqa
def f_name():   return [                                                                            # noqa
                    regex(r"\babs(?=\()"),
                    regex(r"\bacos(?=\()"),
                    regex(r"\bacosh(?=\()"),
                    regex(r"\basin(?=\()"),
                    regex(r"\basinh(?=\()"),
                    regex(r"\batan(?=\()"),
                    regex(r"\batan2(?=\()"),
                    regex(r"\batanh(?=\()"),
                    regex(r"\bcbrt(?=\()"),
                    regex(r"\bceil(?=\()"),
                    regex(r"\bcopysign(?=\()"),
                    regex(r"\bcos(?=\()"),
                    regex(r"\bcosh(?=\()"),
                    regex(r"\bdrem(?=\()"),
                    regex(r"\berf(?=\()"),
                    regex(r"\berfc(?=\()"),
                    regex(r"\bexp(?=\()"),
                    regex(r"\bexpm1(?=\()"),
                    regex(r"\bfact(?=\()"),
                    regex(r"\bfinite(?=\()"),
                    regex(r"\bfloat(?=\()"),
                    regex(r"\bfloor(?=\()"),
                    regex(r"\bfmod(?=\()"),
                    regex(r"\bif(?=\()"),
                    regex(r"\bimodf(?=\()"),
                    regex(r"\bint(?=\()"),
                    regex(r"\bisinf(?=\()"),
                    regex(r"\bisnan(?=\()"),
                    regex(r"\bldexp(?=\()"),
                    regex(r"\bln(?=\()"),
                    regex(r"\blog(?=\()"),
                    regex(r"\blog10(?=\()"),
                    regex(r"\blog1p(?=\()"),
                    regex(r"\bmax(?=\()"),
                    regex(r"\bmin(?=\()"),
                    regex(r"\bmodf(?=\()"),
                    regex(r"\bpow(?=\()"),
                    regex(r"\bremainder(?=\()"),
                    regex(r"\brint(?=\()"),
                    regex(r"\bround(?=\()"),
                    regex(r"\bnearbyint(?=\()"),
                    regex(r"\bsin(?=\()"),
                    regex(r"\bsinh(?=\()"),
                    # regex(r"\bsize(?=\()"),
                    regex(r"\bsqrt(?=\()"),
                    # regex(r"\bsum(?=\()"),
                    # regex(r"\bSum(?=\()"),
                    regex(r"\btan(?=\()"),
                    regex(r"\btanh(?=\()"),
                ]


expr_grammar = ParserPython(expr, reduce_tree=True)



# Copyright (C) 2022-2025 Daniel Billotte, Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from hvcc.types.IR import IRSignalList

from .expr.expr_c_writer import ExprCWriter
from .HeavyObject import HeavyObject


class SignalExpr(HeavyObject):
    """ Handles the math objects.
        Note: obj_eval_functions dict accumulates until reset!
    """

    preamble = "cExprSig"
    obj_eval_functions: Dict = {}

    @classmethod
    def handles_type(cls, obj_type: str) -> bool:
        """ Returns true if the object type can be handled by this class
        """
        return obj_type == "_expr~"

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvMath.h"}

    @classmethod
    def get_C_class_header_code(cls, obj_type: str, args: Dict) -> List[str]:
        eval_funcs = ",\n\t\t".join(cls.obj_eval_functions.values())
        fptr_type = f"{cls.preamble}_evaluator"
        lines = [
            f"typedef void(*{fptr_type})(hv_bInf_t*, hv_bOutf_t);",
            f"{fptr_type} {cls.preamble}_evaluators[{len(cls.obj_eval_functions)}] = {{\n\t\t{eval_funcs}\n\t}};",
        ]
        return lines

    @classmethod
    def get_C_obj_header_code(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        lines = []
        func_name = f"{cls.preamble}_{obj_id}_evaluate"
        cls.obj_eval_functions[obj_id] = func_name
        lines.extend([
            f"static inline void {func_name}(hv_bInf_t* bIns, hv_bOutf_t bOut);",
        ])
        return lines

    @classmethod
    def get_C_obj_impl_code(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        """ (Per object) this creates the _sendMessage function that other objects use to
            send messages to this object.
        """

        lines = []
        expr = args["expressions"][0]
        expr_parser = ExprCWriter(expr)
        expr_lines = expr_parser.to_c_simd("bIns", "bOut")[:-1]
        expr_line = "\n".join(
            [f"\t{line}" for line in expr_lines]
        )
        num_buffers = expr_parser.num_simd_buffers()
        buffer_declaration = "\t// no extra buffers needed"
        if num_buffers > 0:
            buffers = ", ".join([f"Bf{i}" for i in range(0, num_buffers)])
            buffer_declaration = f"\thv_bufferf_t {buffers};"

        func_name = f"Heavy_{{{{name}}}}::{cls.preamble}_{obj_id}_evaluate"
        lines.extend([
            "",
            f"void {func_name}(hv_bInf_t* bIns, hv_bOutf_t bOut) {{",

            buffer_declaration,
            expr_line,
            "}",

        ])
        return lines

    @classmethod
    def get_C_process(cls, process_dict: IRSignalList, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        input_args = []
        for b in process_dict.inputBuffers:
            buf = HeavyObject._c_buffer(b)
            input_args.append(f"VIf({buf})")
        out_buf = HeavyObject._c_buffer(process_dict.outputBuffers[0])
        out_buf = f"VOf({out_buf})"

        call = [
            "",
            "\t// !!! declare this buffer once outside the loop",
            f"\thv_bInf_t input_args_{obj_id}[{args['num_inlets']}] = {{{', '.join(input_args)}}};",
            f"\t{cls.preamble}_evaluators[{len(cls.obj_eval_functions)}](input_args_{obj_id}, {out_buf});"
            "",
        ]

        return call



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2026 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import decimal
import os
import re

from collections import Counter
from collections import OrderedDict
from pathlib import Path
from typing import Optional, Type, Any, Generator

from .HeavyObject import HeavyObject
from .HeavyGraph import HeavyGraph              # pre-converted Heavy graphs
from .HvSwitchcase import HvSwitchcase          # __switchcase
from .PdAudioIoObject import PdAudioIoObject    # adc~/dac~
from .PdBinopObject import PdBinopObject        # binary arithmatic operators
from .PdExprObject import PdExprObject          # expr/expr~
from .PdGraph import PdGraph                    # canvas
from .PdLetObject import PdLetObject            # inlet/inlet~/outlet/outlet~
from .PdMessageObject import PdMessageObject    # msg
from .PdPackObject import PdPackObject          # pack
from .PdReceiveObject import PdReceiveObject    # r/r~/receive/receive~/catch~
from .PdRouteObject import PdRouteObject        # route
from .PdSelectObject import PdSelectObject      # select/sel
from .PdSendObject import PdSendObject          # s/s~/send/send~/throw~
from .PdTriggerObject import PdTriggerObject    # trigger/t
from .PdTableObject import PdTableObject        # table
from .PdUnpackObject import PdUnpackObject      # unpack
from .PdLibSignalGraph import PdLibSignalGraph  # pd/lib abstraction connection checks

from .NotificationEnum import NotificationEnum


class PdParser:

    # library search paths
    __LIB_DIR = Path(Path(__file__).parent, "libs")
    __HVLIB_DIR = Path(Path(__file__).parent, "libs", "heavy")
    __HVLIB_CONVERTED_DIR = Path(Path(__file__).parent, "libs", "heavy_converted")
    __PDLIB_DIR = Path(Path(__file__).parent, "libs", "pd")
    __ELSELIB_DIR = Path(Path(__file__).parent, "libs", "else")
    __CYCLONE_DIR = Path(Path(__file__).parent, "libs", "cyclone")
    __PDLIB_CONVERTED_DIR = Path(Path(__file__).parent, "libs", "pd_converted")

    # detect a dollar argument in a string
    RE_DOLLAR = re.compile(r"\$(\d+)")

    # detect width parameter e.g. "#X obj 172 79 t b b, f 22;"
    RE_WIDTH = re.compile(r", f \d+$")

    # split arguments on non-escaped spaces e.g. "test\ ing"
    RE_SPACE = re.compile(r'(?<!\\)\ ')

    def __init__(self) -> None:
        # the current global value of $0
        # Note(joe): set a high starting value to avoid potential user naming conflicts
        self.__DOLLAR_ZERO = 1000

        # a counter of all Pd objects in the graph
        self.obj_counter: Counter = Counter()

        # search paths at this graph level
        self.search_paths: list[Path] = []

    @classmethod
    def get_supported_objects(cls) -> list:
        """ Returns a set of all pd objects names supported by the parser.
        """
        pd_objects = [os.path.splitext(f)[0] for f in os.listdir(cls.__PDLIB_DIR) if f.endswith(".pd")]
        pd_objects += [f"else/{os.path.splitext(f)[0]}" for f in os.listdir(cls.__ELSELIB_DIR) if f.endswith(".pd")]
        pd_objects += [f"cyclone/{os.path.splitext(f)[0]}" for f in os.listdir(cls.__CYCLONE_DIR) if f.endswith(".pd")]
        pd_objects.extend(cls.__PD_CLASSES.keys())
        return pd_objects

    @classmethod
    def __get_hv_args(cls, pd_path: Path) -> OrderedDict:
        """ Pre-parse the file for Heavy arguments, such that they are available
            as soon as a graph is created.
        """
        num_canvas = -1
        hv_arg_dict = OrderedDict()
        hv_arg_list: Optional[list] = None
        with open(pd_path, "r") as f:
            for li in f:
                if li.startswith("#N canvas"):
                    hv_arg_list = []
                    hv_arg_dict[li.rstrip(";\r\n")] = hv_arg_list
                    num_canvas += 1
                elif "@hv_arg" in li and hv_arg_list is not None:
                    hv_arg_list.append(li.rstrip(";\r\n"))
                elif li.startswith("#X restore"):
                    num_canvas -= 1
                    hv_arg_list = list(hv_arg_dict.values())[num_canvas]
        return hv_arg_dict

    @classmethod
    def get_pd_line(cls, pd_path: Path) -> Generator:
        concat = ""  # concatination state
        with open(pd_path, "r") as f:
            for li in f:
                # concatenate split lines in the Pd file here
                li = li.rstrip("\r\n")  # account for windows CRLF
                if li.endswith(";") and not li.endswith(r"\;"):
                    out = li[:-1]  # remove single ";"
                    if len(concat) > 0:
                        out = f'{concat} {out}'
                        concat = ""  # reset concatenation state
                    yield out
                else:
                    concat = (f'{concat} {li}') if len(concat) > 0 else f'{li}'

    @classmethod
    def split_line(cls, li: str) -> list[str]:
        # remove width parameter
        li = cls.RE_WIDTH.sub("", li)
        # split on non-escaped spaces
        line = cls.RE_SPACE.split(li)
        # replace escaped spaces
        line = [i.replace('\\ ', ' ') for i in line]

        return line

    def add_absolute_search_directory(self, search_dir: Path) -> bool:
        if search_dir.is_dir():
            self.search_paths.append(search_dir)
            return True
        else:
            return False

    def add_relative_search_directory(self, search_dir: Path) -> bool:
        search_dir = Path(
            self.search_paths[0],
            search_dir
        ).absolute()
        return self.add_absolute_search_directory(search_dir)

    def find_abstraction_path(self, local_dir: Path, abs_name: str) -> Optional[Path]:
        """ Finds the full path for an abstraction.
            Checks the local directory first, then all declared paths.
        """

        abs_filename = f"{abs_name}.pd"

        # check local directory first
        abs_path = Path(local_dir.absolute(), abs_filename)
        if abs_path.is_file():
            return abs_path

        # check search paths in reverse order (last added search path first)
        for d in reversed(self.search_paths):
            abs_path = Path(d, abs_filename)
            if abs_path.is_file():
                return abs_path

        return None

    def graph_from_file(
        self,
        file_path: Path,
        obj_args: Optional[list] = None,
        pos_x: int = 0,
        pos_y: int = 0,
        is_root: bool = True,
        pd_graph_class: Type[PdGraph] = PdGraph
    ) -> Any:
        """ Instantiate a PdGraph from a file.
            Note that obj_args does not include $0.
            @param pd_graph_class  The python class to handle specific graph types
        """
        # add main patch directory. The first entry of self.search_paths is
        # assumed to be the root path of the whole system

        if is_root:
            self.search_paths.append(file_path.parent)

        file_hv_arg_dict = self.__get_hv_args(file_path)
        file_iterator = self.get_pd_line(file_path)
        canvas_line: str = file_iterator.__next__()

        self.__DOLLAR_ZERO += 1  # increment $0
        graph_args = [self.__DOLLAR_ZERO] + (obj_args or [])

        if not canvas_line.startswith("#N canvas"):
            g_cls = pd_graph_class(graph_args, file_path, pos_x, pos_y)
            g_cls.add_error(f"Pd files must begin with \"#N canvas\": {canvas_line}")
            return g_cls

        g: PdGraph = self.graph_from_canvas(
            file_iterator,
            file_hv_arg_dict,
            canvas_line,
            graph_args,
            file_path,
            pos_x, pos_y,
            is_root,
            pd_graph_class)

        if is_root:
            if g.get_notices().has_error:
                # return the graph early here as there are already errors and it is
                # clearly invalid, avoids triggering unrelated errors in validation
                return g
            g.validate_configuration()

        return g

    def graph_from_canvas(
        self,
        file_iterator: Generator,
        file_hv_arg_dict: dict[str, list[str]],
        canvas_line: str,
        graph_args: list,
        pd_path: Path,
        pos_x: int = 0,
        pos_y: int = 0,
        is_root: bool = False,
        pd_graph_class: Type[PdGraph] = PdGraph
    ) -> Any:
        """ Instantiate a PdGraph from an existing canvas.
            Note that graph_args includes $0.
            @param file_hv_arg_dict  A dictionary containing all Heavy argument lines
            for each "#N canvas" in this file.
            @param canvas_line  The "#N canvas" which initiates this canvas.
            @param pd_graph_class  The python class to handle specific graph types
        """
        obj_array: Optional[HeavyObject] = None  # an #A (table) object which is currently being parsed

        def finalize_array(array: HeavyObject) -> None:
            declared_size = array.obj_dict["size"]
            values_size = len(array.obj_dict["values"])
            if declared_size != values_size:
                new_size = max(declared_size, values_size)
                array.add_warning(
                    "Table \"{0}\" was declared as having {1} values, "
                    "but {2} were supplied. It will be resized to {3} "
                    "values (any unsupplied values will be zeroed).".format(
                        array.obj_dict["name"],
                        declared_size,
                        values_size,
                        new_size))
                array.obj_dict["size"] = new_size
                if new_size < declared_size:
                    array.obj_dict["values"] = obj_args["values"][:new_size]
                else:
                    array.obj_dict["values"].extend([0.0 for _ in range(new_size - declared_size)])

        g = pd_graph_class(graph_args, pd_path, pos_x, pos_y)

        msg_send: dict = {}
        gui_send: dict = {}
        gui_recv: dict = {}

        # parse and add all Heavy arguments to the graph
        for li in file_hv_arg_dict[canvas_line]:
            line = li.split()
            assert line[4] == "@hv_arg"
            is_required = (line[9] == "true")
            default_value = HeavyObject.force_arg_type(line[8], line[7]) if not is_required else None
            g.add_hv_arg(
                arg_index=int(line[5][2:]) - 1,  # strip off the leading "\$" and make index zero-based
                name=line[6],
                value_type=line[7],
                default_value=default_value,
                required=is_required)

        try:  # this try will capture any critical errors
            for li in file_iterator:
                line = self.split_line(li)

                if line[0] == "#N":
                    if line[1] == "canvas":
                        x = self.graph_from_canvas(
                            file_iterator,
                            file_hv_arg_dict,
                            canvas_line=li,
                            graph_args=graph_args,  # subpatch inherits parent graph arguments, including $0
                            pd_path=pd_path,
                            pos_x=int(line[2]),
                            pos_y=int(line[3]))
                        g.add_object(x)

                elif line[0] == "#X":
                    if line[1] == "restore":
                        if len(line) > 5 and line[5] == "@hv_obj":
                            obj_args = self.__resolve_object_args(
                                obj_args=line[7:],
                                graph=g,
                                raise_on_failure=False,
                                is_root=is_root)
                            if line[6] == "__switchcase":
                                x = HvSwitchcase(
                                    obj_type=line[6],
                                    obj_args=obj_args,
                                    pos_x=int(line[2]),
                                    pos_y=int(line[3]))
                            else:
                                x = HeavyObject(
                                    obj_type=line[6],
                                    obj_args=obj_args,
                                    pos_x=int(line[2]),
                                    pos_y=int(line[3]))
                            return x  # return a Heavy object instead of a graph
                        else:
                            # are we restoring an array object?
                            # do some final sanity checks
                            if obj_array is not None:
                                finalize_array(obj_array)

                                obj_array = None  # done parsing the array

                            # set the subpatch name
                            g.subpatch_name = " ".join(line[5:]) if len(line) > 5 else "subpatch"
                            g = self.__create_send_recv(g, msg_send, gui_send, gui_recv)
                            return g  # pop the graph

                    elif line[1] == "text":
                        # @hv_arg arguments are pre-parsed
                        # TODO(mhroth): is it necessary to split the entire line at once?
                        # always add the comment to the graph, regardless
                        self.obj_counter["text"] += 1
                        g.add_object(HeavyObject(
                            obj_type="comment",
                            obj_args=[" ".join(line[4:])],
                            pos_x=int(line[2]),
                            pos_y=int(line[3])))

                    elif line[1] == "obj":
                        x = None  # a PdObject
                        if len(line) > 4:
                            obj_type = line[4]
                            # sometimes objects have $ arguments in them as well
                            obj_type = self.__resolve_object_args(
                                obj_args=[obj_type],
                                graph=g,
                                is_root=is_root)[0]
                            obj_args = self.__resolve_object_args(
                                obj_args=line[5:],
                                graph=g,
                                is_root=is_root)
                        else:
                            g.add_warning(
                                "This graph contains an empty object. "
                                "It should be removed or defined.",
                                NotificationEnum.WARNING_EMPTY_OBJECT)
                            g.add_object(HeavyObject(
                                obj_type="comment",
                                obj_args=["null object placeholder"],
                                pos_x=int(line[2]),
                                pos_y=int(line[3])))
                            continue

                        if obj_type in ('block~',):
                            # we ignore the object and continue
                            g.add_warning(
                                f"This graph contains a {obj_type} object that is ignored.",
                                NotificationEnum.WARNING_USELESS_OBJECT)
                            g.add_object(HeavyObject(
                                obj_type="comment",
                                obj_args=[f"{obj_type} object placeholder"],
                                pos_x=int(line[2]),
                                pos_y=int(line[3])))
                            continue

                        # do we have an abstraction for this object?
                        abs_path = self.find_abstraction_path(pd_path.parent, obj_type)
                        if abs_path is not None and not g.is_abstraction_on_call_stack(abs_path):
                            # ensure that infinite recursion into abstractions is not possible
                            x = self.graph_from_file(
                                file_path=abs_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False)

                        # is this object in lib/pd_converted?
                        elif Path(self.__PDLIB_CONVERTED_DIR, f"{obj_type}.hv.json").is_file():
                            self.obj_counter[obj_type] += 1
                            hv_path = Path(self.__PDLIB_CONVERTED_DIR, f"{obj_type}.hv.json")
                            x = HeavyGraph(
                                hv_path=hv_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]))

                        # is this object in lib/heavy_converted?
                        elif Path(self.__HVLIB_CONVERTED_DIR, f"{obj_type}.hv.json").is_file():
                            self.obj_counter[obj_type] += 1
                            hv_path = Path(self.__HVLIB_CONVERTED_DIR, f"{obj_type}.hv.json")
                            x = HeavyGraph(
                                hv_path=hv_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]))

                        # is this object in lib/pd?
                        elif Path(self.__PDLIB_DIR, f"{obj_type}.pd").is_file():
                            self.obj_counter[obj_type] += 1
                            pdlib_path = Path(self.__PDLIB_DIR, f"{obj_type}.pd")

                            # mapping of pd/lib abstraction objects to classes
                            # for checking connection validity
                            pd_class = {
                                "abs~": PdLibSignalGraph,
                                "clip~": PdLibSignalGraph,
                                "cos~": PdLibSignalGraph,
                                "dbtopow~": PdLibSignalGraph,
                                "dbtorms~": PdLibSignalGraph,
                                "exp~": PdLibSignalGraph,
                                "ftom~": PdLibSignalGraph,
                                "hip~": PdLibSignalGraph,
                                "lop~": PdLibSignalGraph,
                                "mtof~": PdLibSignalGraph,
                                "powtodb~": PdLibSignalGraph,
                                "q8_rsqrt~": PdLibSignalGraph,
                                "q8_sqrt~": PdLibSignalGraph,
                                "rmstodb~": PdLibSignalGraph,
                                "rsqrt~": PdLibSignalGraph,
                                "sqrt~": PdLibSignalGraph,
                                "wrap~": PdLibSignalGraph
                            }.get(obj_type, PdGraph)

                            x = self.graph_from_file(
                                file_path=pdlib_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False,
                                pd_graph_class=pd_class)

                            # register any object-specific warnings or errors
                            if obj_type in {"rzero~", "rzero_rev~", "czero~", "czero_rev~"}:
                                g.add_warning(
                                    f"[{obj_type}] accepts only signal input. "
                                    "Arguments and control connections are ignored.")

                        # is this object in lib/heavy?
                        elif Path(self.__HVLIB_DIR, f"{obj_type}.pd").is_file():
                            self.obj_counter[obj_type] += 1
                            hvlib_path = Path(self.__HVLIB_DIR, f"{obj_type}.pd")
                            x = self.graph_from_file(
                                file_path=hvlib_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False)

                        # is this object in lib/else?
                        elif Path(self.__ELSELIB_DIR, f"{obj_type}.pd").is_file():
                            self.obj_counter[obj_type] += 1
                            hvlib_path = Path(self.__ELSELIB_DIR, f"{obj_type}.pd")
                            x = self.graph_from_file(
                                file_path=hvlib_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False)

                        # is this object in lib/cyclone?
                        elif Path(self.__CYCLONE_DIR, f"{obj_type}.pd").is_file():
                            self.obj_counter[obj_type] += 1
                            hvlib_path = Path(self.__CYCLONE_DIR, f"{obj_type}.pd")
                            x = self.graph_from_file(
                                file_path=hvlib_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False)

                        # is this object in lib? (sub-directory)
                        elif Path(self.__LIB_DIR, f"{obj_type}.pd").is_file():
                            self.obj_counter[obj_type] += 1
                            hvlib_path = Path(self.__LIB_DIR, f"{obj_type}.pd")
                            x = self.graph_from_file(
                                file_path=hvlib_path,
                                obj_args=obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]),
                                is_root=False)

                        # is this an object that must be programatically parsed?
                        elif obj_type in self.__PD_CLASSES:
                            self.obj_counter[obj_type] += 1
                            obj_class = self.__PD_CLASSES[obj_type]
                            x = obj_class(
                                obj_type,
                                obj_args,
                                pos_x=int(line[2]), pos_y=int(line[3]))

                        elif self.__is_float(obj_type):
                            # parse float literals
                            self.obj_counter["float"] += 1
                            x = HeavyObject(
                                obj_type="__var",
                                obj_args=[float(obj_type)],
                                pos_x=int(line[2]), pos_y=int(line[3]))

                        else:
                            g.add_error(
                                f"Don't know how to parse object \"{obj_type}\". Is it an "
                                "object supported by Heavy? Is it an abstraction? "
                                "Have the search paths been correctly configured?",
                                NotificationEnum.ERROR_UNKNOWN_OBJECT)
                            x = HeavyObject(
                                obj_type="comment",
                                obj_args=[f"null object placeholder ({obj_type})"])

                        assert x is not None
                        index = g.add_object(x)

                        # add gui send/receive objects
                        arg_index = {
                            "nbx":       (6, 7),
                            "vsl":       (6, 7),
                            "hsl":       (6, 7),
                            "vradio":    (4, 5),
                            "hradio":    (4, 5),
                            "bng":       (4, 5),
                            "tgl":       (2, 3),
                            "knob":      (5, 6),
                            "else/knob": (5, 6),
                        }

                        if obj_type in arg_index.keys():
                            send_index = arg_index[obj_type][0]
                            recv_index = arg_index[obj_type][1]

                            if obj_args[send_index] != "empty":
                                gui_send[index] = obj_args[send_index]

                            if obj_args[recv_index] != "empty":
                                gui_recv[index] = obj_args[recv_index]

                    elif line[1] in ("floatatom", "symbolatom"):
                        self.obj_counter[line[1]] += 1
                        obj_args = self.__resolve_object_args(
                                obj_args=line[1:],
                                graph=g,
                                is_root=is_root)
                        x = self.graph_from_file(
                            file_path=Path(self.__PDLIB_DIR, f"{line[1]}.pd"),
                            obj_args=obj_args,
                            pos_x=int(line[2]), pos_y=int(line[3]),
                            is_root=False)
                        index = g.add_object(x)

                        # symbolatom is not supported
                        # due to symbol/__var implementation
                        if line[1] == "floatatom":
                            if obj_args[9] != "-":
                                gui_send[index] = obj_args[9]

                            if obj_args[8] != "-":
                                gui_recv[index] = obj_args[8]

                    elif line[1] == "array":
                        # are we still parsing a previous array? finalize it.
                        if obj_array is not None:
                            finalize_array(obj_array)

                        # array names can have dollar arguments in them.
                        # ensure that they are resolved
                        table_def = self.__resolve_object_args(
                            obj_args=[line[2]],
                            graph=g)[0].split(" ")
                        table_name = table_def[0]
                        # check if we need to extern the table
                        if len(table_def) > 1:
                            table_extern = table_def[1] == "@hv_table"
                        else:
                            table_extern = False
                        # Pd encodes arrays with length greater than 999,999 with
                        # scientific notatation (e.g. 1e6) which Python's int() can't parse
                        table_size = int(decimal.Decimal(line[3]))
                        obj_array = HeavyObject(
                            obj_type="table",
                            # ensure that obj_array has its own values instance
                            obj_args=[table_name, table_size, [], table_extern])
                        # TODO(mhroth): position information
                        g.add_object(obj_array)

                    elif line[1] == "msg":
                        self.obj_counter["msg"] += 1
                        msg = PdMessageObject(
                            obj_type="msg",
                            obj_args=[" ".join(line[4:])],
                            pos_x=int(line[2]),
                            pos_y=int(line[3]))

                        index = g.add_object(msg)

                        if len(msg.obj_dict) > 0:
                            msg_send[index] = []

                            for remote in msg.obj_dict["remote"]:
                                msg_send[index].append(remote)

                    elif line[1] == "connect":
                        g.add_parsed_connection(
                            from_index=int(line[2]),
                            from_outlet=int(line[3]),
                            to_index=int(line[4]),
                            to_inlet=int(line[5]))

                    elif line[1] == "declare":
                        if not is_root:
                            g.add_warning(
                                "[declare] objects are not supported in abstractions. "
                                "They can only be in the root canvas.")
                        elif len(line) >= 4 and line[2] == "-path":
                            pd_parent = pd_path.parent
                            pd_search = Path(pd_parent, line[3])
                            did_add = self.add_relative_search_directory(pd_search)
                            if not did_add:
                                g.add_warning(
                                    f"\"{line[3]}\" is not a valid relative abstraction "
                                    "search path. It will be ignored.")

                        else:
                            g.add_warning(
                                "Heavy only supports the -path flag for the "
                                "declare object.",
                                NotificationEnum.WARNING_DECLARE_PATH)

                    elif line[1] == "coords":
                        pass  # don't do anything with this command
                    elif line[1] == "f":
                        # A line such as "#X f 123"
                        # is found after a "#X restore" when the [pd afasdfasd]
                        # object's box has been resized
                        pass  # don't do anything with this command

                    else:
                        g.add_error(f"Don't know how to parse line: {' '.join(line)}")

                elif line[0] == "#A" and obj_array is not None:
                    # Test that we have an array continuation and extend the array values.
                    try:
                        float(line[1])
                        obj_array.obj_dict["values"].extend([float(f) for f in line[2:] if f != ""])
                    except ValueError:
                        continue

                else:
                    g.add_error(f"Don't know how to parse line: {' '.join(line)}")

        except Exception as e:
            # bubble the Exception back to the root graph where the graph
            # will be returned
            if not g.is_root:
                raise e
            else:
                # NOTE(mhroth): should the exception be added as an error?
                # Sometimes it's all that we have, so perhaps it's a good idea.
                g.add_error(str(e), NotificationEnum.ERROR_EXCEPTION)

        g = self.__create_send_recv(g, msg_send, gui_send, gui_recv)

        return g

    @classmethod
    def __resolve_object_args(
        cls,
        obj_args: list,
        graph: PdGraph,
        raise_on_failure: bool = True,
        is_root: bool = False
    ) -> Any:
        """ Resolve object arguments against the given graph arguments.
            By default this function raises an Exception if it cannot resolve an
            argument.
            This behaviour may be disabled, in which case the unresolved argument
            is replaced with None (which is an otherwise invalid value). A value of
            None typically indicates to a HeavyObject that the default value
            may be used.
        """
        # TODO(mhroth): can this be done more elegantly?
        resolved_obj_args = list(obj_args)  # make a copy of the original obj_args
        for i, a in enumerate(obj_args):
            for m in set(cls.RE_DOLLAR.findall(a)):
                assert isinstance(a, str)
                x = int(m)  # the dollar index (i.e. $x)
                if len(graph.obj_args) > x:
                    a = a.replace(fr"\${m}", str(graph.obj_args[x]))

                # check if hv_args can be used to supply a default value
                elif len(graph.hv_args) > (x - 1):  # heavy args are zero-indexed
                    if not graph.hv_args[x - 1]["required"]:
                        a = a.replace(fr"\${m}", str(graph.hv_args[x - 1]["default"]))
                    else:
                        graph.add_error(
                            f"There is a missing required argument named \"{graph.hv_args[x - 1]['name']}\".",
                            NotificationEnum.ERROR_MISSING_REQUIRED_ARGUMENT)

                elif is_root:
                    # NOTE(mhroth): this case is questionable, but since Pd
                    # defaults to this behavior without warning, so will we.
                    # graph.add_warning(
                    #     "${0} in \"{1}\" in the top-level graph is resolved to "
                    #     "\"0\". It is recommended that you remove $-arguments "
                    #     "from the top-level graph.".format(m, a))
                    a = a.replace(fr"\${m}", "0")

                else:
                    if raise_on_failure:
                        # NOTE(mhroth): this case is questionable, but since Pd
                        # defaults to this behavior without warning, so will we.
                        # graph.add_warning(
                        #     "Object [{0}] requires argument \"{1}\" but the parent "
                        #     "patch does not provide one ({2}). A default value of "
                        #     "\"0\" will be used.".format(obj_type, a, graph.obj_args))
                        a = a.replace(fr"\${m}", "0")
                    else:
                        a = None  # indicate that this argument could not be resolved by replacing it with None
            resolved_obj_args[i] = a
        return resolved_obj_args

    # a mapping of Pd objects to the classes that will parse them
    __PD_CLASSES = {
        "adc~": PdAudioIoObject,
        "dac~": PdAudioIoObject,
        "expr": PdExprObject,
        "expr~": PdExprObject,
        "inlet": PdLetObject,
        "inlet~": PdLetObject,
        "outlet": PdLetObject,
        "outlet~": PdLetObject,
        "pack": PdPackObject,
        "pd": PdGraph,
        "route": PdRouteObject,
        "sel": PdSelectObject,
        "select": PdSelectObject,
        "t": PdTriggerObject,
        "trigger": PdTriggerObject,
        "table": PdTableObject,
        "unpack": PdUnpackObject,
        "s": PdSendObject,
        "send": PdSendObject,
        "s~": PdSendObject,
        "send~": PdSendObject,
        "throw~": PdSendObject,
        "r": PdReceiveObject,
        "receive": PdReceiveObject,
        "r~": PdReceiveObject,
        "receive~": PdReceiveObject,
        "catch~": PdReceiveObject
    }

    # fill in as much of __PD_CLASSES programmatically as possible
    for o in PdBinopObject.get_supported_objects():
        __PD_CLASSES[o] = PdBinopObject

    @classmethod
    def __is_float(cls, x: int) -> bool:
        """ Returns True if the input can be converted to a float. False otherwise.
        """
        try:
            float(x)
            return True
        except Exception:
            return False

    def __create_send_recv(
        self,
        g: PdGraph,
        msg_send: dict,
        gui_send: dict,
        gui_recv: dict
    ) -> PdGraph:
        # parse remote messages
        for index in msg_send.keys():
            first_msg = g.get_object(index)
            conns = first_msg.get_inlet_connections()

            for remote in msg_send[index]:
                self.obj_counter["msg"] += 1
                msg = PdMessageObject("msg", [" ".join(msg for msg in remote["message"])])
                msg_index = g.add_object(msg)

                self.obj_counter["send"] += 1
                send = PdSendObject("send", [remote["receiver"]])
                send_index = g.add_object(send)

                # connect new message to upstream objects of first message
                for conn in conns["0"]:
                    up_obj = conn.from_obj
                    up_index = g.get_objects().index(up_obj)
                    g.add_parsed_connection(up_index, 0, msg_index, 0)

                g.add_parsed_connection(msg_index, 0, send_index, 0)

        # parse gui sends
        for index in gui_send.keys():
            self.obj_counter["send"] += 1
            send = PdSendObject('send', gui_send[index].split())
            send_index = g.add_object(send)

            g.add_parsed_connection(index, 0, send_index, 0)

        # parse gui receives
        for index in gui_recv.keys():
            self.obj_counter["receive"] += 1
            recv = PdReceiveObject('receive', gui_recv[index].split())
            recv_index = g.add_object(recv)

            g.add_parsed_connection(recv_index, 0, index, 0)

        return g



# Copyright (C) 2022-2025 Daniel Billotte, Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Literal, Union

from arpeggio import ParsingExpression as ParseTreeNode, Terminal  # type: ignore
from .expr_arpeggio_grammar import expr_grammar

ParseExpr = Union[ParseTreeNode, Terminal]
NodeType = Literal["binary", "bound_var", "num_i", "num_f", "var", "func", "unary"]


class ExprNode:
    def __init__(
        self,
        node_type: NodeType,
        value: str,
        nodes: list = []
    ) -> None:
        self.type = node_type
        self.value = value
        self.nodes: list[ExprNode] = nodes

    def __str__(self, depth: int = 0) -> str:
        this = f"{'  '*depth}{self.type}:{self.value}\n"
        for node in self.nodes:
            this += node.__str__(depth + 1)
        return this


class ExprArpeggioParser():
    """
    These two methods are the "public" API
    """
    @classmethod
    def parse(cls, expr) -> ParseExpr:
        """Parse the input expression and return a parse tree"""
        return expr_grammar.parse(expr)

    @classmethod
    def parse_to_ast(cls, expr: str) -> ExprNode:
        """Parse the input expression and return an AST"""
        return cls.to_expr_tree(cls.parse(expr))

    """
    Helper methods below
    """
    @classmethod
    def to_expr_tree(cls, expr: ParseExpr) -> ExprNode:
        if expr.rule_name in ("num_i", "num_f", "var"):
            return ExprNode(expr.rule_name, expr.value)
        elif expr.rule_name == "func":
            return ExprNode("func", expr[0].value, [cls.to_expr_tree(p) for p in expr[1:]])
        elif expr.rule_name == "unary":
            return ExprNode("unary", expr[0].value, [cls.to_expr_tree(expr[1])])
        # elif expr.rule_name in ():
        #     # Note: this rule is not used by any expression
        #     # this is RtoL associativity
        #     val = None
        #     tmp = None
        #     subtree = None
        #     for i in range(len(expr)):
        #         if i % 2 == 0:
        #             val = cls.to_expr_tree(expr[i])
        #         else:
        #             foo = ExprNode("binary", str(expr[i]), [val])
        #             if subtree is None:
        #                 subtree = foo
        #             else:
        #                 assert tmp
        #                 tmp.nodes.append(foo)
        #             tmp = foo
        #     assert val and tmp and subtree
        #     tmp.nodes.append(val)
        #     return subtree
        elif expr.rule_name in (
            "lor", "land", "bor", "xor", "band", "eq", "gtlte", "gtlt", "shift", "factor", "term"
        ):
            # this is LtoR associativity
            subtree = None
            for i in range(len(expr)):
                if i % 2 == 0:
                    if subtree is None:
                        subtree = cls.to_expr_tree(expr[i])
                    else:
                        subtree.nodes.append(cls.to_expr_tree(expr[i]))
                else:
                    subtree = ExprNode("binary", str(expr[i]), [subtree])
            assert subtree
            return subtree
        else:
            raise ValueError(f"Unknown rule {expr.rule_name}")



# Heavy Compiler Collection
# Copyright (C) 2024-2026 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

import jinja2
import os
import shutil
import time

from typing import Optional
from pathlib import Path

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.compiler import Generator, CompilerResp, CompilerNotif, CompilerMsg, ExternInfo
from hvcc.types.meta import Meta

from ..copyright import copyright_manager


class c2fmod(Generator):
    """Generates a FMOD plugin wrapper for a given patch."""

    @classmethod
    def compile(
        cls,
        c_src_dir: Path,
        out_dir: Path,
        externs: ExternInfo,
        patch_name: str,
        patch_meta: Meta = Meta(),
        num_input_channels: int = 0,
        num_output_channels: int = 0,
        copyright: Optional[str] = None,
        verbose: Optional[bool] = False,
    ) -> CompilerResp:
        tick = time.time()

        patch_name = patch_name.lower() if patch_name is not None else "heavy"

        in_parameter_list = externs.parameters.inParam
        out_parameter_list = externs.parameters.outParam
        event_list = externs.events.inEvent
        out_event_list = externs.events.outEvent
        table_list = externs.tables

        copyright_c = copyright_manager.get_copyright_for_c(copyright)

        templates_dir = Path(Path(__file__).parent, "templates")
        is_source_plugin = num_input_channels == 0

        out_dir = Path(out_dir, "fmod")
        if not out_dir.exists():
            out_dir.mkdir(parents=True)

        env = jinja2.Environment()
        env.loader = jinja2.FileSystemLoader(
            encoding="utf-8-sig", searchpath=[templates_dir]
        )

        try:

            patch_src_dir = Path(out_dir, "include", "Heavy")
            if patch_src_dir.exists():
                shutil.rmtree(patch_src_dir)
            shutil.copytree(c_src_dir, patch_src_dir)

            heavy_src_files = [f for f in os.listdir(c_src_dir) if f.endswith(".c") or f.endswith(".cpp")]

            src_ext_list = ["h", "hpp", "c", "cpp", "js", "md", "txt"]

            for f in env.list_templates(extensions=src_ext_list):
                file = Path(f)
                file_dir = Path(out_dir, file.parent)

                file_name = file.name.replace("{{name}}", patch_name)
                file_path = Path(file_dir, file_name)

                if not file_path.parent.exists():
                    file_path.parent.mkdir(parents=True)

                with open(file_path, "w") as g:
                    g.write(env.get_template(f).render(
                        name=patch_name,
                        in_params=in_parameter_list,
                        out_params=out_parameter_list,
                        out_events=out_event_list,
                        events=event_list,
                        tables=table_list,
                        pool_sizes_kb=externs.memoryPoolSizesKb,
                        is_source_plugin=is_source_plugin,
                        num_input_channels=num_input_channels,
                        num_output_channels=num_output_channels,
                        heavy_src_files=heavy_src_files,
                        copyright=copyright_c))

            return CompilerResp(
                stage="c2fmod",
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )

        except Exception as e:
            return CompilerResp(
                stage="c2fmod",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    warnings=[],
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )



#!/usr/bin/env python3

import argparse
import json
import re
from pathlib import Path


def _to_symbol_stem(nam_path: Path) -> str:
    stem = nam_path.stem.strip()
    if not stem:
        return "NamModel"

    # Split on non-alphanumeric boundaries, then form PascalCase.
    raw_parts = [p for p in re.split(r"[^A-Za-z0-9]+", stem) if p]
    if not raw_parts:
        return "NamModel"

    parts = []
    for part in raw_parts:
        if part[0].isdigit():
            part = f"M{part}"
        parts.append(part[0].upper() + part[1:])

    return "".join(parts)


def _format_float(value: float) -> str:
    # Keep enough precision for model weights while staying compact.
    return f"{format(value, '.17g')}f"


def convert_nam_to_header(nam_file: Path, header_file: Path) -> tuple[str, int]:
    with nam_file.open("r", encoding="utf-8") as f:
        payload = json.load(f)

    if "weights" not in payload or not isinstance(payload["weights"], list):
        raise ValueError("Input .nam file must contain a 'weights' array")

    weights = payload["weights"]
    for idx, value in enumerate(weights):
        if not isinstance(value, (int, float)):
            raise ValueError(f"weights[{idx}] is not numeric")

    symbol_stem = _to_symbol_stem(nam_file)
    count_name = f"{symbol_stem}WeightsCount"
    weights_name = f"{symbol_stem}Weights"

    lines = [
        "#pragma once",
        "",
        f"static constexpr unsigned int {count_name} = {len(weights)};",
        f"static constexpr float {weights_name}[{count_name}] = {{",
    ]

    row = []
    for idx, value in enumerate(weights, start=1):
        row.append(_format_float(float(value)))
        if idx % 6 == 0:
            lines.append("    " + ", ".join(row) + ",")
            row = []

    if row:
        lines.append("    " + ", ".join(row) + ",")

    lines.append("};")
    lines.append("")

    header_file.parent.mkdir(parents=True, exist_ok=True)
    header_file.write_text("\n".join(lines), encoding="utf-8")

    return weights_name, len(weights)


def main() -> None:
    parser = argparse.ArgumentParser(
        description=(
            "Convert a Neural Amp Modeler .nam JSON file into a C++ header "
            "with static constexpr weight symbols."
        )
    )
    parser.add_argument("input_nam", type=Path, help="Path to input .nam file")
    parser.add_argument("output_header", type=Path, help="Path to output .h file")
    args = parser.parse_args()

    if not args.input_nam.is_file():
        raise FileNotFoundError(f"Input file not found: {args.input_nam}")

    weights_name, count = convert_nam_to_header(args.input_nam, args.output_header)
    print(f"Generated {args.output_header} with {count} weights in symbol {weights_name}.")


if __name__ == "__main__":
    main()


# Heavy Compiler Collection
# Copyright (C) 2025-2026 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

from enum import IntEnum
from typing import Literal, Optional, Union

from pydantic import BaseModel
from pydantic_extra_types.color import Color

from hvcc.version import VERSION


class Coords(BaseModel):
    x: int = 0
    y: int = 0


class Size(BaseModel):
    x: int = 0
    y: int = 0


class Font(IntEnum):
    courier = 0
    helvetica = 1
    times = 2


class LabelShow(IntEnum):
    never = 0
    always = 1
    active = 2
    typing = 3


class LabelPos(IntEnum):
    left = 0
    right = 1
    top = 2
    bottom = 3


class Base(BaseModel):
    type: str
    position: Coords
    size: Size


class BaseUI(Base):
    id: str = ""


class BaseParam(Base):
    parameter: str


class Label(BaseModel):
    text: str
    position: Coords
    color: Color
    font: Font
    font_size: int


class Comment(BaseUI):
    type: Literal["comment"] = "comment"
    text: str
    width: Optional[int] = 0


class Canvas(BaseUI):
    type: Literal["canvas"] = "canvas"
    label: Optional[Label] = None
    bg_color: Color


class Bang(BaseParam):
    type: Literal["bang"] = "bang"
    label: Optional[Label] = None
    flash_time: int
    fg_color: Color
    bg_color: Color


class Toggle(BaseParam):
    type: Literal["toggle"] = "toggle"
    label: Optional[Label] = None
    fg_color: Color
    bg_color: Color
    non_zero: float


class Radio(BaseParam):
    label: Optional[Label] = None
    fg_color: Color
    bg_color: Color
    options: int


class VRadio(Radio):
    type: Literal["vradio"] = "vradio"


class HRadio(Radio):
    type: Literal["hradio"] = "hradio"


class Slider(BaseParam):
    label: Optional[Label] = None
    min: float
    max: float
    fg_color: Color
    bg_color: Color
    logarithmic: bool
    steady: bool


class VSlider(Slider):
    type: Literal["vslider"] = "vslider"


class HSlider(Slider):
    type: Literal["hslider"] = "hslider"


class Knob(BaseParam):
    type: Literal["knob"] = "knob"
    label_size: int
    label_pos: Coords
    label_show: LabelShow
    min: float
    max: float
    fg_color: Color
    bg_color: Color
    init_val: float
    ang_range: int
    ang_offset: int
    log_mode: Literal["lin", "log", "exp"]
    exp_fact: float
    discrete: bool
    ticks: bool
    steps: int
    circular: bool
    jump: bool
    square: bool
    arc_color: Color
    arc_start: float
    arc_show: bool


class Number(BaseParam):
    type: Literal["number"] = "number"
    label: Optional[Label] = None
    fg_color: Color
    bg_color: Color
    log_mode: bool
    log_height: int
    min: float
    max: float


class Float(BaseParam):
    type: Literal["float"] = "float"
    font_height: int
    label_text: str
    label_pos: LabelPos
    min: float
    max: float


GUIObjects = Union[Bang, Toggle, Radio, Slider, Knob, Number, Float, Comment, Canvas]


class Theme(BaseModel):
    obj_corner_radius: Optional[float] = None
    cnv_color: Optional[Color] = None
    text_color: Optional[Color] = None
    io_color: Optional[Color] = None
    bg_color: Optional[Color] = None
    sel_color: Optional[Color] = None
    out_color: Optional[Color] = None


class GraphBase(BaseModel):
    id: str = "mainPatch"
    objects: list[GUIObjects]
    graphs: list["Graph"]


class Graph(GraphBase):
    position: Coords
    gop_start: Coords
    gop_size: Size


class GraphRoot(GraphBase):
    size: Size
    theme: Optional[Theme] = Theme()
    version: str = VERSION



# Copyright (C) 2022-2025 Daniel Billotte, Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import List, Union

from .expr_arpeggio_parser import ExprArpeggioParser, ExprNode, ParseExpr


class BufferAllocator:
    """ Class for managing the swapping of buffers from
        output to input in successive calls.
    """
    def __init__(self) -> None:
        self._avail: set = set()
        self._next: int = 0

    def next(self) -> int:
        """ If a buffer is available return it, otherwise
            allocate a new one and return it.
        """
        if len(self._avail) > 0:
            return self._avail.pop()
        nxt = self._next
        self._next += 1
        return nxt

    def free(self, n: int) -> None:
        """ Return a buffer back to the pool to be reused
        """
        self._avail.add(n)

    def num_allocated(self) -> int:
        """ Return the buffers allocated in so far.
        """
        return self._next


class ExprCWriter:
    def __init__(self, expression: str):
        self.expression = expression
        self.expr_tree: ExprNode = ExprArpeggioParser.parse_to_ast(expression)
        self.parse_tree: ParseExpr = ExprArpeggioParser.parse(expression)
        self.num_buffers: Union[int, None] = None

    def to_ast(self) -> ExprNode:
        return self.expr_tree

    def to_parse_tree(self) -> ParseExpr:
        return self.parse_tree

    def to_c_simd(
        self,
        vv_in: Union[str, None] = None,
        v_out: str = ""
    ) -> List[str]:
        self._simd_bind_variables(vv_in)
        self._simd_replace_constants()
        return self._to_c_simd(v_out)

    def num_simd_buffers(self) -> int:
        if self.num_buffers is None:
            return len(self.to_c_simd())
        return self.num_buffers

    def to_c_nested(self) -> str:
        return self._to_c_nested()

    def _simd_replace_constants(self) -> None:
        self._simd_replace_constants_R(self.expr_tree)

    def _simd_replace_constants_R(self, tree: ExprNode) -> None:
        if tree.type == "func" and tree.value.startswith("_load_"):
            return  # don't re-replace anything

        for i in range(len(tree.nodes)):
            node = tree.nodes[i]
            if node.type == "num_i":
                tree.nodes[i] = ExprNode("func", "_load_i", [node])
            elif node.type == "num_f":
                tree.nodes[i] = ExprNode("func", "_load_f", [node])
            else:
                self._simd_replace_constants_R(node)

    def _simd_bind_variables(self, a_name: Union[str, None] = "A") -> None:
        assert a_name
        self._bind_vars_R(self.expr_tree, a_name)

    def _bind_vars_R(self, tree: ExprNode, a_name: str) -> None:
        if tree.type == "var":
            self._bind_var_node(tree, a_name)
        else:
            for node in tree.nodes:
                self._bind_vars_R(node, a_name)

    def _bind_var_node(self, node: ExprNode, a_name) -> None:
        if node.type != "var":
            pass
        parts = re.match(r"\$(v|f)(\d+)", node.value)
        assert parts
        node.value = f"{a_name}[{int(parts[2])-1}]"
        node.type = "bound_var"

    def _to_c_simd(self, v_out: str) -> List[str]:
        ba = BufferAllocator()
        lines: List[str] = []

        def _to_c_simd_R(expr_tree: ExprNode, r_vec: Union[str, None] = None) -> str:
            if expr_tree.value == "-" and len(expr_tree.nodes) == 1:
                # Unary minus, treat as (0-arg) by inserting a 0-const node.
                zero_node = ExprNode("func", "_load_f", [ExprNode("num_f", "0.0f")])
                expr_tree.nodes.insert(0, zero_node)

            if expr_tree.type in ("num_i", "num_f", "var", "bound_var"):
                return expr_tree.value

            args = []
            buffers = []

            if (
                expr_tree.type == "func"
                and expr_tree.value.startswith("_load_")
            ):
                next_buf = f"Bf{ba.next()}"
                args.append(f"&{next_buf}")
                const_value = _to_c_simd_R(expr_tree.nodes[0])
                for i in range(8):
                    args.append(const_value)
            else:
                for node in expr_tree.nodes:
                    val = _to_c_simd_R(node)
                    args.append(val)
                    if isinstance(val, str) and val.startswith("Bf"):
                        buffers.append(val)
                if r_vec is not None:
                    next_buf = r_vec
                    args.append(next_buf)
                else:
                    next_buf = f"Bf{ba.next()}"
                    args.append(f"&{next_buf}")

            f_name = ExprOpMap.get_hv_func_simd(expr_tree.value)
            lines.append(f"{f_name}({', '.join(args)});")
            for b in buffers:
                ba.free(int(b[2]))

            return next_buf

        lines.append(_to_c_simd_R(self.expr_tree, v_out))
        self.num_buffers = ba.num_allocated()
        return lines

    def _to_c_nested(self) -> str:
        """Output C-code as nested function calls"""

        def _to_c_nested_R(expr_tree: ExprNode) -> str:
            if expr_tree.type in ("num_i", "num_f", "var"):
                return expr_tree.value
            else:
                f_name = ExprOpMap.get_hv_func(expr_tree.value)
                args = [_to_c_nested_R(p) for p in expr_tree.nodes]
                return f"{f_name}({', '.join(args)})"

        return f"{_to_c_nested_R(self.expr_tree)};"


class ExprOpMap:
    op_map = {
        "~":            "hv_bit_not_f",
        "!":            "hv_not_f",
        "*":            "hv_mul_f",
        "/":            "hv_div_f",
        "%":            "hv_modulo_f",
        "+":            "hv_add_f",
        "-":            "hv_sub_f",
        "<":            "hv_lt_f",
        "<=":           "hv_lte_f",
        ">":            "hv_gt_f",
        ">=":           "hv_gte_f",
        "<<":           "hv_shl_f",
        ">>":           "hv_shr_f",
        "==":           "hv_eq_f",
        "!=":           "hv_neq_f",
        "^":            "hv_exc_or_f",
        "&":            "hv_bit_and_f",
        "&&":           "hv_log_and_f",
        "|":            "hv_bit_or_f",
        "||":           "hv_log_or_f",
        "abs":          "hv_abs_f",
        "acos":         "hv_acos_f",
        "acosh":        "hv_acosh_f",
        "asin":         "hv_asin_f",
        "asinh":        "hv_asinh_f",
        "atan":         "hv_atan_f",
        "atan2":        "hv_atan2_f",
        "atanh":        "hv_atanh_f",
        # "avg":          "hv_?_f",           # averages all elements of a table
        # "Avg":          "hv_?_f",           # averages elements of a specified boundary of a table
        "cbrt":         "hv_cbrt_f",
        "ceil":         "hv_ceil_f",
        "copysign":     "hv_copysign_f",
        "cos":          "hv_cos_f",
        "cosh":         "hv_cosh_f",
        "erf":          "hv_erf_f",
        "erfc":         "hv_erfc_f",
        "exp":          "hv_exp_f",
        "expm1":        "hv_expm1_f",
        "fact":         "hv_fact_f",
        "finite":       "hv_finite_f",
        "float":        "hv_cast_if_expr",
        "floor":        "hv_floor_f",
        "fmod":         "hv_fmod_f",
        "if":           "hv_if_f",
        "imodf":        "hv_imod_f",
        "int":          "hv_cast_fi_expr",
        "isinf":        "hv_isinf_f",
        "isnan":        "hv_isnan_f",
        "ldexp":        "hv_ldexp_f",
        "ln":           "hv_ln_f",
        "log":          "hv_log_f",
        "log10":        "hv_log10_f",
        "log1p":        "hv_log1p_f",
        "max":          "hv_max_f",
        "min":          "hv_min_f",
        "modf":         "hv_modf_f",
        "pow":          "hv_pow_f",
        "remainder":    "hv_remainder_f",
        "rint":         "hv_rint_f",        # round to nearest int
        "round":        "hv_round_f",       # round to nearest int
        "nearbyint":    "hv_rint_f",        # round to nearest int
        "sin":          "hv_sin_f",
        "sinh":         "hv_sinh_f",
        # "size":         "hv_?_f",          # size of a table
        "sqrt":         "hv_sqrt_f",
        # "sum":          "hv_?_f",          # sum of all elements of a table
        # "Sum":          "hv_?_f",          # sum of elemnets of a specified boundary of a table
        "tan":          "hv_tan_f",
        "tanh":         "hv_tanh_f",
        "_load_f":      "hv_var_k_f",
        "_load_i":      "hv_var_k_i",
    }

    @classmethod
    def get_hv_func(cls, symbol: str) -> str:
        return cls.op_map[symbol]

    @classmethod
    def get_hv_func_simd(cls, symbol: str) -> str:
        return f"__{cls.op_map[symbol]}"
